"""
Manages logging. There are two parts: messages and performance statistics

Messages are structured as follows:
    activity: some kind of significant processing
    result: outcome of the major processing
    status: status of an activity: start, finish
    exception: an exception occurred
    error: a processing error occurred

Perofrmance logging is provided as well. Usage is:
    blockGuid = logger.startBlock("My block")
    ... lots of code ...
    logger.endBlock(blockGuid)
    ... more code ...
    print(logger.performanceDF)  # performance data
"""
from SBstoat import _helpers

import pandas as pd
import time

SEPARATOR = "/"  # Separates levels in performance logs
LEVEL_ACTIVITY = 1
LEVEL_RESULT = 2
LEVEL_STATUS = 3
LEVEL_EXCEPTION = 4
LEVEL_ERROR = 5
LEVEL_DETAILS = 6
LEVEL_MAX = LEVEL_DETAILS
# Dataframe columns
COUNT = "count"
MEAN = "mean"
TOTAL = "total"
# Attributes used in equals comparisons
STATISTIC_ATTR_MERGE = ["count", "total"]
STATISTIC_ATTR_EQUALS = list(STATISTIC_ATTR_MERGE)
STATISTIC_ATTR_EQUALS.append("mean")
LOGGER_ATTR = ["isReport", "toFile", "startTime", "logLevel", "unpairedBlocks",
      "blockDct", "performanceDF", "statisticDct"]


class BlockSpecification():
    # Describes an entry for timing a block of code
    guid = 0

    def __init__(self, block):
        self.guid = BlockSpecification.guid
        BlockSpecification.guid += 1
        self.startTime = time.time()
        self.block = block
        self.duration = None  # Time duration of the block

    def setDuration(self):
        self.duration = time.time() - self.startTime

    def __repr__(self):
        result = "guid: %d, block: %s, startTime: %f"  \
              % (self.guid, self.block, self.startTime)
        if self.duration is not None:
            result += ", duration: %f" % self.duration
        return result


class Statistic():
    # Statistics for a block
    def __init__(self, block=None):
        self.block = block
        self.count = 0
        self.total = 0.0
        self.mean = None

    def __repr__(self):
        result = "Statistic[block: %s, count: %d, total: %f"  \
              % (self.block, self.count, self.total)
        if self.mean is not None:
            result += "mean: %f"  % self.mean
        result += "]"
        return result

    def copy(self):
        return _helpers.copyObject(self)

    def update(self, value):
        self.count += 1
        self.total += value

    def equals(self, other):
        true = True
        for attr in STATISTIC_ATTR_EQUALS:
            result = self.__getattribute__(attr) == other.__getattribute__(attr)
            if not isinstance(result, bool):
                result = all(result)
            true = true and result
        return true

    def merge(self, other):
        newStatistic = self.copy()
        for attr in STATISTIC_ATTR_MERGE:
            value = self.__getattribute__(attr) + other.__getattribute__(attr)
            newStatistic.__setattr__(attr, value)
        return newStatistic

    def summarize(self):
        if self.count == 0:
            self.mean = 0.0
        else:
            self.mean = self.total/self.count


class Logger():

    def __init__(self, isReport=True, toFile=None, logLevel=LEVEL_ERROR,
          logPerformance=False):
        self.isReport = isReport
        self.toFile = toFile
        self.logPerformance = logPerformance
        self.startTime = time.time()
        self.logLevel = logLevel
        self.unpairedBlocks = 0  # Count of blocks begun without an end
        self.blockDct = {}  # key: guid, value: BlockSpecification. Must be lock protected.
        self._performanceDF = None  # Summarizes performance report
        self.statisticDct = {}

    def copy(self):
        return _helpers.copyObject(self)

    def equals(self, other):
        true = True
        for attr in LOGGER_ATTR:
            result = self.__getattribute__(attr) == other.__getattribute__(attr)
            if not isinstance(result, bool):
                result = all(result)
            true = true and result
        return true

    @property
    def performanceDF(self):
        """
        Summarizes the performance data collected.

        Returns
        -------
        pd.Series
            index: block name
            Columns: COUNT, MEAN
        """
        if self._performanceDF is None:
            # Accumulate the durations
            self.unpairedBlocks = len(self.blockDct)
            #
            indices = list(self.statisticDct.keys())
            _ = [s.summarize() for s in self.statisticDct.values()]
            means = [self.statisticDct[b].mean for b in indices]
            counts = [self.statisticDct[b].count for b in indices]
            totals = [self.statisticDct[b].total for b in indices]
            self._performanceDF = pd.DataFrame({
                  COUNT: counts,
                  MEAN: means,
                  TOTAL: totals,
                  })
            self._performanceDF.index = indices
            self._performanceDF = self.performanceDF.sort_index()
        return self._performanceDF

    def formatPerformanceDF(self, numLevel=2):
        """
        Constructs a string that trims the index of performanceDF

        Parameters
        ----------
        numLevel: int
            Number of identifier separators included.

        Returns
        -------
        str
        """
        performanceDF = self.performanceDF.copy()
        indices = self.performanceDF.index
        newIndices = list(indices)
        for idx, item in enumerate(indices):
            splits = [s for s in item.split(SEPARATOR) if len(s) > 0]
            if len(splits) > numLevel - 1:
                newIndices[idx] = ".." + SEPARATOR +  \
                       SEPARATOR.join(splits[-numLevel:])
        performanceDF.index = newIndices
        return str(performanceDF)

    def getFileDescriptor(self):
        if self.toFile is not None:
            return open(self.toFile, "a")
        return None

    @staticmethod
    def join(*args):
        """
        Joins together a list of block names.

        Parameters
        ----------
        *args: list-str

        Returns
        -------
        str
        """
        if len(args) == 1:
            return args[0]
        if len(args) == 2:
            if len(args[0]) == 0:
                return args[1]
        return SEPARATOR.join(args)

    def _write(self, msg, numNL):
        relTime = time.time() - self.startTime
        newLineStr = ('').join(["\n" for _ in range(numNL)])
        newMsg = "\n%s%f: %s" % (newLineStr, relTime, msg)
        if self.toFile is None:
            print(newMsg)
        else:
            with open(self.toFile, "a") as fd:
                fd.write(newMsg)

    def activity(self, msg):
        # Major processing activity
        if self.isReport and (self.logLevel >= LEVEL_ACTIVITY):
            self._write("***%s***" %msg, 2)

    def result(self, msg):
        # Result of an activity
        if self.isReport and (self.logLevel >= LEVEL_RESULT):
            self._write("\n **%s" %msg, 1)

    def status(self, msg):
        # Progress message
        if self.isReport and (self.logLevel >= LEVEL_STATUS):
            self._write("    (%s)" %msg, 0)

    def exception(self, msg):
        # Progress message
        if self.isReport and (self.logLevel >= LEVEL_EXCEPTION):
            self._write("    (%s)" %msg, 0)

    def error(self, msg, excp):
        # Progress message
        if self.isReport and (self.logLevel >= LEVEL_ERROR):
            fullMsg = "%s: %s" % (msg, str(excp))
            self._write("    (%s)" % fullMsg, 0)

    def details(self, msg):
        # Progress message
        if self.isReport and (self.logLevel >= LEVEL_DETAILS):
            self._write("    (%s)" %msg, 0)

    ###### BLOCK TIMINGS ######
    def startBlock(self, block:str)->float:
        """
        Records the beginning of a block.

        Parameters
        ----------
        block: name of the block

        Returns
        -------
        int: identifier for the BlockSpecification
        """
        if self.logPerformance:
            spec = BlockSpecification(block)
            self.blockDct[spec.guid] = spec
            return spec.guid
        return None

    def _merge(self, other):
        """
        Merges the results of another logger.
        """
        newLogger = self.copy()
        for block, statistic in self.statisticDct.items():
            otherStatistic = other.statisticDct[block]
            newLogger.statisticDct[block] = statistic.merge(otherStatistic)
        return newLogger

    @staticmethod
    def merge(others):
        if len(others) == 0:
            raise ValueError("Cannot provide an empty list.")
        newLogger = others[0]
        curLogger = newLogger
        for other in others[1:]:
            newLogger = curLogger._merge(other)
            curLogger = newLogger
        return newLogger


    def endBlock(self, guid:int):
        """
        Records the end of a block. Items are removed as
        statistics are accumulated.

        Parameters
        ----------
        guid: identifies the block instance
        """
        if self.logPerformance:
            if not guid in self.blockDct.keys():
                self.exception("missing guid: %d" % guid)
            else:
                spec = self.blockDct[guid]
                spec.setDuration()
                if not spec.block in self.statisticDct.keys():
                    self.statisticDct[spec.block] = Statistic(block=spec.block)
                self.statisticDct[spec.block].update(spec.duration)
                del self.blockDct[spec.guid]

    def update(self, other):
        """
        Copies the content of other to this object.

        Parameters
        ----------
        other: Logger
        """
        _ = _helpers.copyObject(other, self)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 14:24:09 2020

@author: hsauro
@author: joseph-hellerstein

A NamedTimeseries is a container of multiple vairables that are obtained at the
same timepoints. Variables can be accessed by time using ("[", "]").
Various properties of the common timepoints can be accessed, such as start
and end.

Usage:
   # Create from file
   timeseries = NamedTimeseries(csvPath="myfile.txt")
   print(timeseries)  # dispaly a tabular view of the timeseries
   # NamedTimeseries can use len function
   length = len(timeseries)  # number of rows
   # Extract the numpy array values using indexing
   timeValues = timeseries["time"]
   s1Values = timeseries["S1"]
   # Get the start and end times
   startTime = timeseries.start
   endTime = timeseries.end
   # Create a new time series that subsets the variables of the old one
   colnames = ["time", "S1", "S2"]
   newTimeseries = mkNamedTimeseries(colnames, timeseries[colnames])
   # Create a new timeseries that excludes time 0
   ts2 = timeseries[1:]
   # Create a new column variable
   timeseries["S8"] = timeseries["time"]**2 + 3*timeseries["S1"]
   timeseries["S9"] = 10  # Assign a constant to all rows
"""

from SBstoat import rpickle

import copy
import csv
import numpy as np
import pandas as pd

DELIMITER = ","
TIME = "time"


################## FUNCTIONS ########################
def mkNamedTimeseries(*args):
    """
    Constructs a NamedTimeseries from different arguments.

    Parameters
    ----------
    args: list
        [list-str, np.ndarray]
        [NamedTimeseries]
        [str]
    """
    if len(args) == 2:
        return NamedTimeseries(colnames=args[0], array=args[1])
    if len(args) == 1:
        if isinstance(args[0], NamedTimeseries):
            return args[0]
        if isinstance(args[0], str):
            return NamedTimeseries(csvPath=args[0])
    raise ValueError("Specification for NameTimeseries: %s"
          % str(args))


def arrayEquals(arr1, arr2):
    """
        Tests equality of two numpy arrays.

        Parameters
        ---------
        arr1: numpy array
        arr2: numpy array

        Returns
        -------
        boolean
    """
    if np.shape(arr1) != np.shape(arr2):
        return False
    value = sum([np.abs(v1 - v2) for v1, v2 in
        zip(arr1.flatten(), arr2.flatten())])
    return np.isclose(value, 0)

def cleanColnames(colnames):
    """
    Fixes column names by deleting '[', ']'

    Parameters
    ----------
    colnames: list-str

    Returns
    -------
    list-str
    """
    return [s[1:-1] if s[0] == '[' else s for s in colnames]


################## CLASSES ########################
class NamedTimeseries(rpickle.RPickler):

    def __init__(self,
          csvPath=None,
          colnames=None, array=None,
          namedArray=None,
          dataframe=None,
          timeseries=None,
          isNone=False):
        """
        Parameters
        ---------
        csvPath: str
            path to CSV file
        colnames: list-str
        array: np.ndarray
            values corresponding to colnames
        namedArray: NamedArray
        dataframe: pd.DataFrame
            index: time
        timeseries: NamedTimeseries
        isNone: bool
            Initialization to None values

       Usage
       -----
           data = NamedTimeseries(csvPath="mydata.csv")

       Notes
       -----
           At one of the following most be non-None:
             csvPath, colnames & array, dataframe, timeseries
        """
        if isNone:
            return
        if timeseries is not None:
            # Copy the existing object
            for k in timeseries.__dict__.keys():
                self.__setattr__(k,
                      copy.deepcopy(timeseries.__dict__[k]))
        else:
            if csvPath is not None:
                allColnames, self.values = self._load(csvPath)
            elif (colnames is not None) and (array is not None):
                allColnames = colnames
                self.values = array
            elif namedArray is not None:
                allColnames = cleanColnames(namedArray.colnames)
                self.values = namedArray
            elif dataframe is not None:
                if dataframe.index.name == TIME:
                    df = dataframe.reset_index()
                else:
                    df = dataframe
                allColnames = df.columns.tolist()
                self.values = df.to_numpy()
            else:
                msg = "Source should be a file path, colnames & array"
                raise ValueError(msg)
            timeIdxs = [i for i, c in enumerate(allColnames) if c.lower() == TIME]
            if len(timeIdxs) != 1:
                raise ValueError("Must have exactly one time column")
            self.values = np.array(self.values)  # Ensure not NamedArray
            self.values = self.values.astype(float)
            allColnames[timeIdxs[0]] = TIME
            self.allColnames = []  # all column names
            self.colnames = []  # Names of non-time columns
            self._indexDct = {} # index for columns
            _ = [self._addColname(n) for n in allColnames]
            times = self.values[:, self._indexDct[TIME]]
            self.start = min(times)
            self.end = max(times)

    @classmethod
    def rpConstruct(cls):
        """
        Overrides rpickler.rpConstruct to create a method that
        constructs an instance without arguments.

        Returns
        -------
        Instance of cls
        """
        return cls(isNone=True)

    def __repr__(self):
        df = self.to_dataframe()
        return str(df)

    def _addColname(self, name):
        self.allColnames.append(name)
        self._indexDct[name] = self.allColnames.index(name)
        self.colnames = list(self.allColnames)
        if TIME in self.colnames:
            self.colnames.remove(TIME)

    def __setitem__(self, reference, value):
        """
            Assigns a value, and optionally, creates a new columns.
            The value may be 0 or 1 dimenson for a new column.
            If reference is a list of columns, then the dimension
            must be <length of timeseries> X <number of columns in reference>

            Parameters
            ---------
            reference: str or str/list-str
            value: scalar, array
        """
        indices = self._getColumnIndices(reference, isValidate=False)
        if indices is None:
            # New column
            if not isinstance(reference, str):
                raise ValueError("New column must be a string.")
            # New column is being added
            self._addColname(reference)
            numDim = len(np.shape(value))
            numRow = np.shape(self.values)[0]
            if numDim == 0:
                # Make it 1-d
                value = np.repeat(value, numRow)
                numDim = len(np.shape(value))
            if numDim == 1:
                if len(value) != numRow:
                    raise ValueError("New column has %d elements not %d"
                          % (len(value), numRow))
                arr = np.reshape(value, (numRow, 1))
            else:
                msg = "New column must be a scalar or a 1-d array"
                raise ValueError(msg)
            self.values = np.concatenate([self.values, arr],
                  axis=1)
        else:
            # Not a new column
            if isinstance(value, np.ndarray):
                values = value
            else:
                values = np.repeat(value, np.shape(self.values[:, indices]))
            self.values[:, indices] = values

    def _getColumnIndices(self, reference, isValidate=True):
        """
            Returns indices for reference.

            Parameters
            ---------
            reference: str/list-str
            isValidate: boolean
                raise an error if index doesn't exist

            Returns
            -------
            list-int
        """
        try:
            if isinstance(reference, str):
                indices = self._indexDct[reference]
            elif isinstance(reference, list):
                indices = [v for k, v in self._indexDct.items() if k in reference]
            else:
                raise KeyError
        except KeyError:
            if isValidate:
                raise ValueError(
                      "NamedTimeseries invalid index: %s" % reference)
            indices = None
        return indices

    def __delitem__(self, reference):
        if not isinstance(reference, str):
            raise ValueError("del takes a string column names as its argument.")
        idx = self._getColumnIndices(reference)
        self.values = np.delete(self.values, idx, axis=1)
        self.colnames.remove(reference)
        self.allColnames.remove(reference)
        self._indexDct = {n: self.allColnames.index(n)
              for n in self.allColnames}

    def __getitem__(self, reference):
        """
        Returns data for the requested variables or rows.
        Columns are specified by a str or list-str.
        Rows are indicated by an int, list-int, or slice.

        Parameters
        ---------
        reference: str/list-str/slice/list-int/int

        Returns
        -------
        np.array/NamedTimeseries
            np.ndarray for column type reference
            NamedTimeseries for row type reference
        """
        def handleStr(reference):
            indices = self._getColumnIndices(reference)
            return self.values[:, indices]
        # Handle common case first
        if isinstance(reference, str):
            if reference.lower() == TIME:
                reference = TIME
            indices = self._indexDct[reference]
            return self.values[:, indices]
        # Do more complex types
        isInt = False
        isListInt = False
        isSlice = False
        isListStr = False
        if isinstance(reference, list):
            if isinstance(reference[0], str):
                isListStr = True
            elif isinstance(reference[0], int):
                isListInt = True
        elif isinstance(reference, int):
            isInt = True
        elif isinstance(reference, slice):
            isSlice = True
        # Process row type references
        if isListStr:
            timeStrs = [c for c in reference if c.lower() == TIME]
            if len(timeStrs) == 1:
                ref = timeStrs[0]
                idx = reference.index(ref)
                reference.insert(idx, TIME)
                reference.remove(ref)
            elif len(timeStrs) == 0:
                pass
            else:
                raise ValueError("Reference to multiple different time columns.")
            return handleStr(reference)
        if isListInt:
            return NamedTimeseries(colnames=self.allColnames,
                  array=self.values[reference, :])
        if isInt:
            return NamedTimeseries(colnames=self.allColnames,
                  array=self.values[[reference], :])
        if isSlice:
            allIndices = list(range(len(self)))
            indices = allIndices[reference]
            return NamedTimeseries(colnames=self.allColnames,
                  array=self.values[indices, :])
        #
        raise ValueError("Invalid reference to NamedTimeseries.")

    def __len__(self):
        return np.shape(self.values)[0]

    def _load(self, filePath):
        """
        Load data from a CSV file.

        Parameters
        ----------

        filePath: str
              Path to the file containing time series data. Data should be arranged
              in columns, first colums representing time, remaining columns will
              be whatever timeseries data is avaialble.
        selectedVariables : list of strings
              List of names of floating species that are the columns in the
              time series data. If the list of names is missing then it is assumed that
              all the time series columns should be used in the fit

        Returns
        ------
        list-str, array
            list-str: colnames
            array: values
        """
        # Open the data file and look for the header line
        try:
            with open(filePath, 'r') as f:
                reader = csv.reader(f, delimiter=DELIMITER)
                variablesStr = f.readline()
                variablesStr = variablesStr.strip()
                names = variablesStr.split(DELIMITER)
                values = np.array(list(reader)).astype(float)
        except IOError:
            raise ValueError("There is a problem with the data path %s" % filePath)
        if not all([isinstance(v, str) for v in names]):
            raise ValueError('No column headers for file: %s' % filePath)
        #
        return names, values

    def flatten(self):
        """
        Creates a one dimensional array of values

        Returns
        ------
        array
        """
        indices = [self._indexDct[c] for c in self.colnames]
        arr = self.values[:, indices]
        return arr.flatten()

    def selectTimes(self, selectorFunction):
        """
            Selects a subset of rows based on time values by using
            a boolean valued selector function.

            Parameters
            ----------

            selectorFunction: Function
                argument: time value
                returns: boolean

            Returns
            ------
            array
        """
        rowIdxs = [i for i, t in enumerate(self[TIME])
              if selectorFunction(t)]
        colIdxs = [i for i in self._indexDct.values()
              if self._indexDct[TIME] != i]
        array = self.values[rowIdxs, :]
        return array[:, colIdxs]

    def copy(self, isInitialize=False):
        """
        Copies the orginal timeseries with an option for initializing columns
        to 0.

        Parameters
        ----------
        isIinitialize: bool

        Return
        -----
        NamedTimeseries
        """
        ts = copy.deepcopy(self)
        if isInitialize:
            for column in self.colnames:
                ts[column] = 0.0
        return ts

    def equals(self, other):
        """
        Checks if two timeseries have the same columns and values.

        Parameters
        ----------
        other: NamedTimeseries

        Returns
        -------
        bool
        """
        diff = set(self.allColnames).symmetric_difference(other.allColnames)
        if len(diff) > 0:
            return False
        checks = []
        for col in self.allColnames:
            checks.append(arrayEquals(self[col], other[col]))
        return all(checks)

    def to_dataframe(self):
        """
        Creates a pandas dataframe from the NamedTimeseries.

        Returns
        ------
        pd.DataFrame
            columns: self.colnames
            index: Time
        """
        dct = {c: self.values[:, i] for c, i in self._indexDct.items()}
        df = pd.DataFrame(dct)
        return df.set_index(TIME)

    def to_csv(self, path):
        df = self.to_dataframe()
        df = df.reset_index()
        df.to_csv(path, index=False)

    def _getTimeseriesOrList(self, others):
        if isinstance(others, NamedTimeseries):
            others = [others]
        return others

    def concatenateColumns(self, others):
        """
        Creates a NamedTimeseries that is the column concatenation of the current
        with others. Duplicate column names are resolved by adding a "_".

        Parameters
        ----------
        others: single or list-NamedTimeseries

        Returns
        ------
        NamedTimeseries

        Usage
        -----
        newTS = ts.concatenateColumns(ts1, ts2, ts3)
        """
        others = self._getTimeseriesOrList(others)
        dfs = [self.to_dataframe()]
        colnames = list(self.colnames)
        for ts in others:
            if len(ts) != len(self):
                raise ValueError("NamedTimeseries must have the same length")
            df = ts.to_dataframe()
            for col in df.columns:
                if col in colnames:
                    newName = "%s_" % col
                    df = df.rename(columns={col: newName})
                    colnames.append(newName)
            dfs.append(df)
        concatDF = pd.concat(dfs, axis=1)
        return NamedTimeseries(dataframe=concatDF)

    def concatenateRows(self, others):
        """
        Creates a NamedTimeseries that is the concatenation of the current
        rows with others.

        Parameters
        ----------
        others: single or list-NamedTimeseries

        Returns
        ------
        NamedTimeseries

        Usage
        -----
        newTS = ts.concatenateRows(ts1, ts2, ts3)
        """
        others = self._getTimeseriesOrList(others)
        dfs = [self.to_dataframe()]
        for ts in others:
            diff = set(self.allColnames).symmetric_difference(ts.allColnames)
            if len(diff) > 0:
                raise ValueError("NamedTimeseries must have the same columns.")
            df = ts.to_dataframe()
            dfs.append(df)
        concatDF = pd.concat(dfs, axis=0)
        return NamedTimeseries(dataframe=concatDF)

    def _getStringOrListstring(self, arg):
        if isinstance(arg, str):
            return [arg]
        return arg

    def subsetColumns(self, colnames):
        """
        Creates a NamedTimeseries consisting of a subset of columns.

        Parameters
        ----------
        colnames: single column name or list of column names

        Returns
        ------
        NamedTimeseries

        Usage
        -----
        newTS = ts.subsetColumns("S1", "S2")  # The new timeseries only has S1, S2
        """
        colnames = self._getStringOrListstring(colnames)
        df = self.to_dataframe()
        return NamedTimeseries(dataframe=df[colnames])

    def isEqualShape(self, otherTS):
        """
        Verifies that data shape and column names are the same.

        Parameters
        ----------
        otherTS: NamedTimeseries

        Returns
        ------
        boolean
        """
        diff = set(self.colnames).symmetric_difference(otherTS.colnames)
        if len(diff) > 0:
            return False
        return np.shape(self.values) == np.shape(otherTS.values)

    def getTimesForValue(self, column, reference):
        """
        Finds the times at which the column assumes the specified value.
        Does linear interpolation to find times.

        Parameters
        ----------
        column: str
        reference: float
            value that is searched/interpolated

        Returns
        ------
        float
        """
        SMALL_FRAC = 10e-6
        valueArr = self[column]
        timeArr = self[TIME]
        values = [1 if v > reference + SMALL_FRAC else -1 if v < reference - SMALL_FRAC
              else 0 for v in valueArr]
        last = 0
        times = []
        # Look for when cross 0
        for idx, value  in enumerate(values):
            if value == 0:
                # Found an exact match
                times.append(timeArr[idx])
            elif last == 0:
                # Do nothing
                pass
            elif last == value:
                # Does not cross 0
                pass
            elif last != value:
                # Interpolate
                if last < value:
                    lrgIdx = idx
                    smlIdx = idx - 1
                else:
                    lrgIdx = idx - 1
                    smlIdx = idx
                frac = (reference - valueArr[smlIdx])/(
                      valueArr[lrgIdx] - valueArr[smlIdx])
                timeDiff = timeArr[lrgIdx] - timeArr[smlIdx]
                time = timeArr[smlIdx] + frac*timeDiff
                times.append(time)
            else:
                raise RuntimeError("Should not get here")
            last = value
        return times

    def rename(self, curName, newName):
        """
        Creates a new timeseries, renaming the specified column.

        Parameters
        ----------
        curName: str
           Name of the column to change
        newName: str
           New name of the column to change
        """
        if not curName in self._indexDct.keys():
            msg = "Invalid reference to NamedTimeseries. No column %s" % curName
            raise ValueError(msg)
        newTimeseries = self.copy()
        newTimeseries[newName] = self[curName]
        del newTimeseries[curName]
        return newTimeseries



import SBstoat
from SBstoat.modelFitter import ModelFitter
from SBstoat.timeseriesPlotter import TimeseriesPlotter, TIME
from SBstoat.namedTimeseries import NamedTimeseries
from SBstoat._version import __version__
# Constants
METHOD_BOTH = SBstoat._modelFitterCore.METHOD_BOTH
METHOD_DIFFERENTIAL_EVOLUTION = SBstoat._modelFitterCore.METHOD_DIFFERENTIAL_EVOLUTION
METHOD_LEASTSQ = SBstoat._modelFitterCore.METHOD_LEASTSQ



from setuptools import setup
import os.path
import codecs

with open("README.md", "r") as fh:
    long_description = fh.read()

# The following two methods were copied from
# https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            print (line)
            delim = '"' if '"' in line else "'"
            print ('delim = ', delim)
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name='SBstoat',    # This is the name of your PyPI-package.
    packages=['SBstoat'],
    version=get_version("SBstoat/_version.py"),   # Update the version number for new releases
    #scripts=['teutilities'],    # The name of your scipt, and also the command you'll be using for calling it
    author='H M Sauro, J Hellerstein',
    author_email='hsauro@uw.edu',
    url='http://tellurium.analogmachine.org',
    description='Parameter fitting for SBML models',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    classifiers=[
       'License :: OSI Approved :: MIT License',
       'Programming Language :: Python :: 3.6',
       'Programming Language :: Python :: 3.7',
       'Operating System :: OS Independent',
    ],
    install_requires=[
        'lmfit',
        'numpy', 
        'tellurium',
        'pandas',
        'nose',
        'python-libsbml',
        'docstring_expander',
    ],
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-
"""
Created on Aug 14, 2020

@author: joseph-hellerstein

Manages options for a figure of plots.

A figure may contain 1 or more plot, and each plot may contain
1 or more line. An option is applied to a single scope: figure, plot, line.
If the option is singled valued, then it applies to all instances
of the scope. If it has multiple values, then the index of the value
corresponds to the instance within the scope. A figure scope is
always single valued.
"""

from SBstoat._statementManager import StatementManager

from docstring_expander.kwarg import Kwarg
import matplotlib.pyplot as plt


PLOT = "plot"  # identifies a plotting function
EXPAND_KEYPHRASE = "Expansion keyphrase. Expands to help(PlotOptions()). Do not remove. (See timeseriesPlotter.EXPAND_KEYPRHASE.)"
# Options
ALPHA = "alpha"
BINS = "bins"
COLUMNS = "columns"
COLOR = "color"
FIGSIZE = "figsize"
LEGEND = "legend"
LINESTYLE = "linestyle"
MARKER = "marker"
MARKERSIZE = "markersize"
NUM_COL = "numCol"
NUM_ROW = "numRow"
SUBPLOT_WIDTH_SPACE = "subplotWidthSpace"
SUPTITLE = "suptitle"
TIMESERIES2 = "timeseries2"
TITLE = "title"
TITLE_FONTSIZE = "titleFontsize"
TITLE_POSITION = "titlePosition"
XLABEL = "xlabel"
XLIM = "xlim"
XTICKLABELS = "xticklabels"
YLABEL = "ylabel"
YTICKLABELS = "yticklabels"
YLIM = "ylim"


#########################
class ExKwarg(Kwarg):
    """Extends Kwarg with scope information."""
    FIGURE = "figure"
    PLOT = "plot"
    LINE = "line"

    def __init__(self, name:str, scope:str, doc:str=None,
              dtype:type=None, default:object=None):
        self.scope = scope
        super().__init__(name, dtype=dtype, default=default)
        shortScope = "(%s)" % self.scope[0]
        self.doc = "%s %s" % (shortScope, doc)


# Options common to most plots
BASE_OPTIONS = [COLUMNS, COLOR, LEGEND, LINESTYLE, MARKER, NUM_COL, MARKERSIZE,
      NUM_ROW, SUBPLOT_WIDTH_SPACE, SUPTITLE, TIMESERIES2, TITLE,
      TITLE_FONTSIZE, TITLE_POSITION, XLABEL, XLIM, YLABEL, YLIM]
FF = ExKwarg.FIGURE
PP = ExKwarg.PLOT
LL = ExKwarg.LINE
KWARGS = [
      ExKwarg(ALPHA, LL, doc="transparency; in [0, 1]", dtype=float),
      ExKwarg(BINS, PP, doc="number of bins in a histogram plot", dtype=int),
      ExKwarg(COLOR, LL, doc="color of the line", dtype=str, default="blue"),
      ExKwarg(COLUMNS, FF, doc= "List of columns to plot", dtype=list, default=[]),
      ExKwarg(FIGSIZE, FF, doc= "(horizontal width, vertical height)",
      dtype=list, default=[8, 6]),
      ExKwarg(LEGEND, FF, doc= "Tuple of str for legend", dtype=list),
      ExKwarg(LINESTYLE, LL, doc= "Line style", dtype=str),
      ExKwarg(MARKER, LL, doc= "Marker for line", dtype=str),
      ExKwarg(MARKERSIZE, LL, doc="Size of marker for the line; >0",
      dtype=float),
      ExKwarg(NUM_ROW, FF, doc= "rows of plots", dtype=int),
      ExKwarg(NUM_COL, FF, doc= "columns of plots", dtype=int),
      ExKwarg(SUBPLOT_WIDTH_SPACE, FF, doc= "horizontal space between plots",
      dtype=float),
      ExKwarg(TIMESERIES2, FF, doc= "second timeseries"),
      ExKwarg(TITLE, PP, doc= "plot title", dtype=str),
      ExKwarg(TITLE_FONTSIZE, FF, doc= "point size for title", dtype=float),
      ExKwarg(TITLE_POSITION, FF,
      doc= "(x, y) relative position; x,y in [0, 1]; (0,0) is lower left."),
      ExKwarg(SUPTITLE, FF, doc= "Figure title", dtype=str),
      ExKwarg(XLABEL, PP, doc= "x axis title", dtype=str),
      ExKwarg(XLIM, FF, doc= "order pair of lower and upper", dtype=list),
      ExKwarg(XTICKLABELS, FF, doc= "list of labels for x ticks", dtype=list),
      ExKwarg(YLABEL, FF, doc= "label for x axis", dtype=str),
      ExKwarg(YLIM, FF, doc= "order pair of lower and upper", dtype=str),
      ExKwarg(YTICKLABELS, FF, doc= "list of labels for y ticks", dtype=list),
      ]
KWARG_DCT = {k.name: k for k in KWARGS}
AX_STATEMENT_KW = {
      MARKERSIZE: 's',
      }
HEADER= """
A figure may contain 1 or more plot, and each plot may contain
1 or more line. Figure (f), plot (p), and line (c) are the possible
scope of an option.  If the option is singled valued, then it applies
to all instances of its scope. If it has multiple values, then the
index of the value corresponds to the instance within the scope.
A figure scope is always single valued.
"""


#########################
class PlotOptions(object):
    """
    Container for plot options. Common method for activating options for an axis.
    """

    def __init__(self):
        ### PRIVATE
        self._axes = None
        ### PUBLIC
        self.figure = None
        for kwarg in KWARGS:
          self.set(kwarg.name, kwarg.default, isForce=True)

    def __str__(self):
        stg = "Supported plotting options are:\n"
        for kwarg in KWARGS:
          stg += "  %s: %s" % (kwarg.name, kwarg.doc)
        return stg

    def set(self, attribute, value, isOverride=False, isForce=False):
        isOK = False
        if isForce:
            isOK = True
        if attribute in self.__dict__.keys():
            isOK = True
        if isOK:
            if isForce or isOverride  \
                  or (self.__getattribute__(attribute) is None):
                self.__setattr__(attribute, value)
        else:
            raise ValueError("Unknown plot option: %s" % attribute)

    def setDict(self, dct, excludes=[]):
        """
        Changes setting based on attributes, values in a dictionary.
        Only those attributes that are initialized are consumed.

        Parameters
        ----------
        dct: dict
            dictionary of attribute value pairs
        excludes: list-sstr
            attributes not to be set
        """
        for attribute, value in dct.items():
            if not attribute in excludes:
                self.set(attribute, value, isOverride=True)

    def initialize(self, timeseries, numRow=None, numCol=None, **kwargs):
        """
        Initializes information for plotting.

        Parameters
        ---------
        timeseries: NamedTimeseries
            a timeseries to be plotted
        numRow: int
        numCol: int
        kwargs: dict
        """
        if numRow is not None:
            self.numRow = numRow
        if numCol is not None:
            self.numCol = numCol
        self.setDict(kwargs, excludes=[NUM_ROW, NUM_COL])  # Record the options
        if len(self.columns) == 0:
            self.columns = timeseries.colnames
        if self.figsize is None:
             self.figsize = plt.gca().figure.get_size_inches()

    def get(self, name, figIdx=None, plotIdx:int=None, lineIdx:int=None):
        """
        Return the value for this instance of the option.

        Parameters
        ----------
        name: str
            option name
        figIdx: int
            index of the Figure. Always 0.
        plotIdx: int
            index of the plot
        lineIdx: int
            index of the line

        Returns
        -------
        object

        Notes
        -----
        1. Must detect single vs. multi-valued instances.
        """
        # Initialize and validate
        exkwarg = KWARG_DCT[name]
        idxDct = {
              ExKwarg.FIGURE: figIdx,
              ExKwarg.PLOT: plotIdx,
              ExKwarg.LINE: lineIdx,
              }
        # Get the value
        value = self.__getattribute__(name)
        idx = idxDct[exkwarg.scope]
        if idx is None:
            return value
        if value is None:
            return value
        if isinstance(value, exkwarg.dtype):
            return value
        if not isinstance(value, list):
            try:
                # Try to coerce type
                value = exkwarg.dtype(value)
                return value
            except:
                pass
            raise ValueError("Invalid data type for option %s" % name)
        result = value[idxDct[exkwarg.scope]]
        return result

    def do(self, ax, statement:StatementManager=None, **kwargs):
        """
        Enacts options taking into account the scope of the option.
        Note that options related to NUM_ROW, NUM_COL are handled elsewhere.

        Parameters
        ----------
        ax: Matplotlib.Axes
        statement: StatementManager
        kwargs: dict
            arguments for get
        """
        def setLineAttribute(ax, name:str, statement:StatementManager=None,
              **kwargs):
            """
            Updates the statement for the line being plotted.
            """
            if statement is None:
                return
            if "Axes.plot" in str(statement.func):
                # Some options are not alloed for line plots
                if name in [MARKER, MARKERSIZE]:
                    return
            value = self.get(name, **kwargs)
            if value is not None:
                if name in AX_STATEMENT_KW.keys():
                    dct={AX_STATEMENT_KW[name]: value}
                else:
                    dct={name: value}
                statement.addKwargs(**dct)
        #
        def setFigPlotAttribute(ax, name:str, **kwargs):
            value = self.__getattribute__(name)
            if value is not None:
                func = eval("ax.set_%s" % name)
                statement = StatementManager(func)
                statement.addPosarg(self.get(name, **kwargs))
                statement.execute()
        # Attributes processed by this method
        names = [ALPHA, COLOR, LINESTYLE, MARKER, MARKERSIZE, XLABEL, YLABEL,
             XLIM, XTICKLABELS, YLIM, YTICKLABELS]
        for name in names:
            kwarg = KWARG_DCT[name]
            if kwarg.scope == ExKwarg.LINE:
                setLineAttribute(ax, name, statement=statement, **kwargs)
            else:
                setFigPlotAttribute(ax, name, **kwargs)
        if statement is not None:
            statement.execute()
        # Other attributes
        if self.title is not None:
            title = self.get(TITLE, **kwargs)
            titlePosition = self.get(TITLE_POSITION, **kwargs)
            titleFontsize = self.get(TITLE_FONTSIZE, **kwargs)
            statement = StatementManager(ax.set_title)
            statement.addPosarg(title)
            if self.titlePosition is not None:
                statement.addKwargs(x=titlePosition[0])
                statement.addKwargs(y=titlePosition[1])
                statement.addKwargs(transform=ax.transAxes)
            if self.titleFontsize is not None:
                statement.addKwargs(fontsize=titleFontsize)
            statement.execute()
        value = self.__getattribute__(SUBPLOT_WIDTH_SPACE)
        if value is not None:
            plt.subplots_adjust(wspace=value)
        if self.legend is not None:
            ax.legend(self.legend)
        if self.suptitle is not None:
            plt.suptitle(self.suptitle)
